#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include <unistd.h>

#include <pthread.h>
#include <SDL.h>
#include <SDL_ttf.h>

#define nil NULL

void
panic(char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	fprintf(stderr, "error: ");
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
	va_end(ap);
	exit(1);
	SDL_Quit();
}

int makepty(void);
int pty;

#define MARGIN 10
#define TOP MARGIN
#define LEFT MARGIN
#define RIGHT (LEFT + 72*hsp)

#define WIDTH (RIGHT + hsp)

char *fontpath = "teletype.ttf";
SDL_Color white = { 0xFF, 0xFF, 0xFF, 0 };
SDL_Color black = { 0x00, 0x00, 0x00, 0 };

SDL_Renderer *renderer;
TTF_Font *font;
SDL_Texture *textures[2];
int curpaper;
SDL_Texture *paper;
SDL_Rect rect;
int curx;
int hsp, vsp;
int paperheight;
int winheight;

typedef struct Glyph Glyph;
struct Glyph
{
	int w, h;
	SDL_Texture *tex;
};
Glyph glyphs[128];

void
draw(void)
{
	SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0);
	SDL_RenderClear(renderer);
	SDL_RenderCopy(renderer, paper, nil, &rect);
	SDL_RenderPresent(renderer);
}

void
clearpaper(void)
{
	SDL_SetRenderTarget(renderer, paper);
	SDL_SetRenderDrawColor(renderer, 0xFF, 0xFF, 0xFF, 0xFF);
	SDL_RenderClear(renderer);
	SDL_SetRenderTarget(renderer, nil);
}

void
movepaper(int n)
{
	SDL_Rect r;
	SDL_Texture *other;

	curpaper ^= 1;
	other = textures[curpaper];
	SDL_SetRenderTarget(renderer, other);
	SDL_SetRenderDrawColor(renderer, 0xFF, 0xFF, 0xFF, 0xFF);
	SDL_RenderClear(renderer);
	r = rect;
	r.y = -n;
	SDL_RenderCopy(renderer, paper, nil, &r);
	SDL_SetRenderTarget(renderer, nil);
	paper = other;
}

void
printchar(char c)
{
	SDL_Rect r;
	c &= 0177;
	SDL_SetRenderTarget(renderer, paper);
	r.x = curx;
	r.y = paperheight - MARGIN - vsp;
	r.w = glyphs[c].w;
	r.h = glyphs[c].h;
	SDL_RenderCopy(renderer, glyphs[c].tex, nil, &r);
	SDL_SetRenderTarget(renderer, nil);
	curx += hsp;
}

void
recvchar(char c)
{
//	printf("%o\n", c);
	int n;

	if(c >= '!' && c <= '_')
		printchar(c);
	else switch(c){
	case ' ':
		curx += hsp;
		break;
	case '\b':
		curx -= hsp;
		break;
	case '\r':
		curx = LEFT;
		break;
	case '\n':
		movepaper(vsp);
		break;
	case '\t':
		n = (curx-LEFT)/hsp;
		n = (n+8)&~7;
		curx = LEFT+n*hsp;
		break;
	}
	if(curx < LEFT) curx = LEFT;
	if(curx+hsp > RIGHT) curx = RIGHT-hsp;
}

void
keydown(SDL_Keysym keysym)
{
	char c;
	int ctrl, shift;
	int key = keysym.sym;
	ctrl = keysym.mod & KMOD_CTRL;
	shift = keysym.mod & KMOD_SHIFT;
	if(key <= 0177){

		/* This defines our keyboard!
		 * Map to the characters the ASR-33 actually has,
		 * or return if the key doesn't exist. */
		if(islower(key))
			key = toupper(key);
		switch(key){
		case '-': key = ':'; break;
		case '=': key = '-'; break;
		case '`': key = 0175; break;	// ALT MODE
		case '\'': return;
		case '[': return;
		case ']': return;
		case '\\': key = '\n'; break;
		case '\b': key = 0177; break;
		case '\t': return;
		}


		if(shift){
			/* not all keys can be shifted (they'll be stuck) */
			if(key >= 'A' && key <= 'J' || key >= 'Q' && key <= 'Z')
				return;
			key ^= 020;
		}
		if(ctrl)
			key &= ~0100;

		c = key;
		write(pty, &c, 1);
//		recvchar(key);
	}
}

void
makefont(void)
{
	SDL_Surface *surf;
	char *chars = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_";
	char *s;
	char tmp[2];
	Glyph *g;

	tmp[1] = '\0';
	for(s = chars; *s; s++){
		g = &glyphs[*s];
		tmp[0] = *s;
		surf = TTF_RenderText_Blended(font, tmp, black);
		g->w = surf->w;
		g->h = surf->h;
		g->tex = SDL_CreateTextureFromSurface(renderer, surf);
		SDL_FreeSurface(surf);
	}
}

void
resize(int h)
{
	printf("%d\n", h);
	winheight = h;
	rect.y = winheight-paperheight;;
}

void
winevent(SDL_WindowEvent ev)
{
	switch(ev.event){
	case SDL_WINDOWEVENT_RESIZED:
		resize(ev.data2);
		break;
	}
}

Uint32 evtype;

void*
readthread(void *p)
{
	char c;
	SDL_Event ev;

	memset(&ev, 0, sizeof(SDL_Event));
	ev.type = evtype;

	while(1){
		read(pty, &c, 1);

		write(1, &c, 1);
		ev.user.code = c;
		SDL_PushEvent(&ev);
	}
}

int
main(int argc, char *argv[])
{
	SDL_Window *window;
	SDL_Event event;
	int running;
	pthread_t thr;

	if(TTF_Init() < 0)
		panic("Couldn't initialize TTF: %s\n",SDL_GetError());
	font = TTF_OpenFont(fontpath, 18);
	if(font == nil)
		panic("Couldn't load font %s: %s\n", fontpath, SDL_GetError());
	TTF_SetFontKerning(font, 0);
	TTF_SetFontHinting(font, TTF_HINTING_MONO);

	curx = LEFT;
	hsp = 13;
	vsp = 19;
	winheight = 24*vsp + 2*MARGIN;

	if(SDL_CreateWindowAndRenderer(WIDTH, winheight, 0, &window, &renderer) < 0)
		panic("SDL_CreateWindowAndRenderer() failed: %s\n", SDL_GetError());

	evtype = SDL_RegisterEvents(1);

	pty = makepty();

	makefont();

	paperheight = winheight*2;
	textures[0] = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET, WIDTH, paperheight);
	textures[1] = SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_TARGET, WIDTH, paperheight);
	paper = textures[curpaper];
	rect.x = 0;
	rect.y = winheight-paperheight;
	rect.w = WIDTH;
	rect.h = paperheight;
	clearpaper();

	draw();

	pthread_create(&thr, nil, readthread, nil);

	running = 1;
	while(running){
		if(SDL_WaitEvent(&event) < 0)
			panic("SDL_PullEvent() error: %s\n", SDL_GetError());
		switch(event.type){
		case SDL_MOUSEBUTTONDOWN:
			break;

		case SDL_WINDOWEVENT:
			winevent(event.window);
			break;

		case SDL_KEYDOWN:
			keydown(event.key.keysym);
			break;
		case SDL_QUIT:
			running = 0;
			break;

		default:
			if(event.type == evtype)
				recvchar(event.user.code);
			break;
		}
		draw();
	}
	return 0;
}
